import { __decorate, __extends } from 'tslib';
import { InjectionToken, Inject, Injectable, Injector, Pipe, isDevMode, NgModule, Optional } from '@angular/core';
import 'rxjs/add/operator/toPromise';
import { ReduxActionDispatcher, ReduxStateDecorator, ReduxReducerDecorator, ReduxStateDecoratorForClass } from '@harmowatch/redux-decorators';
export { ReduxActionContextDecoratorForClass as ReduxActionContext, ReduxActionDecoratorForMethod as ReduxAction, ReduxReducerDecoratorForMethod as ReduxReducer, ReduxStateDecoratorForClass as ReduxState } from '@harmowatch/redux-decorators';
import { AsyncSubject } from 'rxjs/AsyncSubject';
import { Observable } from 'rxjs/Observable';
import 'rxjs/add/operator/map';
import { ReplaySubject } from 'rxjs/ReplaySubject';
import 'rxjs/add/operator/distinctUntilChanged';
import 'rxjs/add/operator/takeWhile';
import { BehaviorSubject } from 'rxjs/BehaviorSubject';
import { createStore } from 'redux';
import { CommonModule } from '@angular/common';

var ReduxStateDefinitionToken = /** @class */ (function (_super) {
    __extends(ReduxStateDefinitionToken, _super);
    function ReduxStateDefinitionToken() {
        return _super.call(this, 'ReduxStateDefinitionToken') || this;
    }
    return ReduxStateDefinitionToken;
}(InjectionToken));
var ReduxStore = /** @class */ (function (_super) {
    __extends(ReduxStore, _super);
    function ReduxStore() {
        return _super.call(this, 'ReduxStore') || this;
    }
    return ReduxStore;
}(InjectionToken));
var ReduxRegistry = /** @class */ (function () {
    function ReduxRegistry(store) {
        if (store === void 0) { store = null; }
        ReduxRegistry.reset();
        ReduxRegistry.registerStore(store);
    }
    ReduxRegistry.reset = function () {
        ReduxRegistry._store = new AsyncSubject();
    };
    ReduxRegistry.registerStore = function (store) {
        ReduxRegistry.reset();
        ReduxRegistry._store.next(store);
        ReduxRegistry._store.complete();
        ReduxActionDispatcher.dispatchedActions.subscribe(function (action) {
            var reduxAction = {
                type: action.type,
                payload: action.payload,
            };
            store.dispatch(reduxAction);
            if (action.onDispatchSuccess) {
                action.onDispatchSuccess();
            }
        });
    };
    ReduxRegistry.registerState = function (state) {
        ReduxRegistry.getStore().then(function (store) {
            var stateConfig = ReduxStateDecorator.get(state.constructor);
            var initialState = state.getInitialState();
            Promise
                .resolve(initialState instanceof Observable ? initialState.toPromise() : initialState)
                .then(function (initialValue) {
                store.dispatch({
                    payload: {
                        initialValue: initialValue,
                        name: stateConfig.name,
                    },
                    type: ReduxRegistry.ACTION_REGISTER_STATE,
                });
            });
        });
    };
    ReduxRegistry.getStore = function () {
        return new Promise(ReduxRegistry._store.subscribe.bind(ReduxRegistry._store));
    };
    return ReduxRegistry;
}());
ReduxRegistry.ACTION_REGISTER_STATE = "@harmowatch/ngx-redux-core/registerState";
ReduxRegistry._store = new AsyncSubject();
ReduxRegistry.decorators = [
    { type: Injectable },
];
ReduxRegistry.ctorParameters = function () { return [
    { type: undefined, decorators: [{ type: Inject, args: [ReduxStore,] },] },
]; };
var ReduxSelector = /** @class */ (function (_super) {
    __extends(ReduxSelector, _super);
    function ReduxSelector(selector, stateProvider) {
        if (selector === void 0) { selector = '/'; }
        var _this = this;
        if (!selector.startsWith(ReduxSelector.DELIMITER) && !stateProvider) {
            throw new Error('You need to provide a state provider, if you use relative selectors');
        }
        _this = _super.call(this, 1) || this;
        ReduxRegistry.getStore().then(function (store) {
            var next = function () {
                _this.next(ReduxSelector.getValueByState(store.getState(), selector, stateProvider));
            };
            store.subscribe(function () { return next(); });
            next();
        });
        return _this;
    }
    ReduxSelector.normalize = function (selector, stateProvider) {
        if (!selector.startsWith(ReduxSelector.DELIMITER)) {
            var stateName = ReduxStateDecorator.get(stateProvider).name;
            return "" + ReduxSelector.DELIMITER + stateName + ReduxSelector.DELIMITER + selector;
        }
        return selector;
    };
    ReduxSelector.getValueByState = function (state, selector, stateProvider) {
        var value = ReduxSelector.normalize(selector, stateProvider).split(ReduxSelector.DELIMITER)
            .filter(function (propertyKey) { return propertyKey !== ''; })
            .reduce(function (previousValue, propertyKey) {
            if (!previousValue || !previousValue.hasOwnProperty(propertyKey)) {
                return null;
            }
            return previousValue[propertyKey];
        }, (state));
        return value;
    };
    return ReduxSelector;
}(ReplaySubject));
ReduxSelector.DELIMITER = '/';
var ReduxStateProvider = /** @class */ (function () {
    function ReduxStateProvider(stateDefs) {
        if (stateDefs === void 0) { stateDefs = []; }
        var _this = this;
        this.selectorCache = {};
        var _a = (ReduxStateDecorator.get(this.constructor) || {}).name, name = _a === void 0 ? null : _a;
        if (!name) {
            throw new Error('Unable to resolve state name! Make sure you\'ve decorated the provider by "@ReduxState"!');
        }
        this.name = ReduxStateDecorator.get(this.constructor).name;
        this.stateDef = stateDefs.find(function (def) { return ReduxStateDecorator.get(def.provider).name === name; });
        if (!this.stateDef) {
            throw new Error('Unable to resolve state definition! Make sure you\'ve registered the provider to ReduxModule!');
        }
        this.reducerMethodsByType = (this.stateDef.reducers || [])
            .map(function (clazz) { return _this.getReducerMethods(new clazz()); })
            .reduce(function (all, curr) { return [].concat(curr, all); }, [])
            .reduce(function (methodsByType, reducer) {
            var type = ReduxActionDispatcher.getType(reducer.type);
            return Object.assign({}, methodsByType, (_a = {}, _a[type] = [reducer.method].concat(methodsByType[type] || []), _a));
            var _a;
        }, {});
        ReduxStateProvider.instancesByName[this.name] = this;
    }
    ReduxStateProvider.prototype.getInitialState = function () {
        throw new Error('Method "getInitialState" not implemented.');
    };
    ReduxStateProvider.prototype.select = function (selector) {
        if (selector === void 0) { selector = ''; }
        var stateType = (this.constructor);
        selector = ReduxSelector.normalize(selector, stateType);
        if (!this.selectorCache[selector]) {
            this.selectorCache[selector] = new ReduxSelector(selector, stateType).distinctUntilChanged();
        }
        return (this.selectorCache[selector]);
    };
    ReduxStateProvider.prototype.getState = function () {
        var _this = this;
        return ReduxRegistry.getStore().then(function (store) {
            return ReduxSelector.getValueByState(store.getState(), '/' + _this.name);
        });
    };
    ReduxStateProvider.prototype.reduce = function (state, action) {
        var reducerMethods = this.reducerMethodsByType[action.type] || [];
        return reducerMethods.reduce(function (stateToReduce, method) { return method(stateToReduce, action); }, state);
    };
    ReduxStateProvider.prototype.getReducerMethods = function (reducerClassInstance) {
        return Object.values(Object.getPrototypeOf(reducerClassInstance))
            .map(function (method) {
            return {
                method: ((method)).bind(reducerClassInstance),
                type: ReduxReducerDecorator.get(method),
            };
        })
            .filter(function (reducer) { return reducer && reducer.type; })
            .reduce(function (all, curr) { return all.concat([].concat(curr.type).map(function (type) { return (Object.assign({}, curr, { type: type })); })); }, []);
    };
    return ReduxStateProvider;
}());
ReduxStateProvider.instancesByName = {};
ReduxStateProvider.ctorParameters = function () { return [
    { type: Array, decorators: [{ type: Inject, args: [ReduxStateDefinitionToken,] },] },
]; };
function ReduxSelect(expression, context) {
    return function (target, propertyKey) {
        var stateName = ReduxStateDecorator.get(context).name;
        Object.defineProperty(target, propertyKey, {
            get: function () { return ReduxStateProvider.instancesByName[stateName].select(expression); }
        });
    };
}
var ReduxReducerProvider = /** @class */ (function () {
    function ReduxReducerProvider() {
        this.stateProviders = {};
    }
    Object.defineProperty(ReduxReducerProvider.prototype, "rootReducer", {
        get: function () {
            return this.reduce.bind(this);
        },
        enumerable: true,
        configurable: true
    });
    ReduxReducerProvider.prototype.addStateProvider = function (provider) {
        if (!this.stateProviders[provider.name]) {
            ReduxRegistry.registerState(provider);
            this.stateProviders[provider.name] = provider;
        }
        else {
            throw new Error("State \"" + provider.name + "\" is registered twice! Make sure your state name is unique!");
        }
    };
    ReduxReducerProvider.prototype.reduce = function (rootState, action) {
        if (action.type === ReduxRegistry.ACTION_REGISTER_STATE) {
            var regAction = (((action)));
            return Object.assign({}, rootState, (_a = {}, _a[regAction.payload.name] = regAction.payload.initialValue, _a));
        }
        return Object.values(this.stateProviders).reduce(function (stateToReduce, provider) {
            return Object.assign({}, stateToReduce, (_a = {}, _a[provider.name] = provider.reduce(stateToReduce[provider.name], action), _a));
            var _a;
        }, rootState);
        var _a;
    };
    return ReduxReducerProvider;
}());
ReduxReducerProvider.decorators = [
    { type: Injectable },
];
ReduxReducerProvider.ctorParameters = function () { return []; };
var ReduxSelectPipe = /** @class */ (function () {
    function ReduxSelectPipe(stateDefs, injector) {
        if (stateDefs === void 0) { stateDefs = []; }
        this.provider = (injector.get(stateDefs[0].provider));
    }
    ReduxSelectPipe.prototype.transform = function (selector) {
        return this.provider.select(selector);
    };
    return ReduxSelectPipe;
}());
ReduxSelectPipe.decorators = [
    { type: Pipe, args: [{ name: 'reduxSelect' },] },
];
ReduxSelectPipe.ctorParameters = function () { return [
    { type: Array, decorators: [{ type: Inject, args: [ReduxStateDefinitionToken,] },] },
    { type: Injector, },
]; };
var ReduxTestingStore = /** @class */ (function () {
    function ReduxTestingStore() {
        this.state = new BehaviorSubject(null);
    }
    ReduxTestingStore.factory = function () {
        return new ReduxTestingStore();
    };
    ReduxTestingStore.prototype.setState = function (state, value) {
        var _this = this;
        var name = ReduxStateDecorator.get(state).name;
        var nextState = Object.assign({}, this.state.getValue(), (_a = {}, _a[name] = value, _a));
        this.state.next(nextState);
        return this.state
            .takeWhile(function (currentState) { return currentState !== nextState; })
            .toPromise().then(function () { return _this.state.getValue(); });
        var _a;
    };
    ReduxTestingStore.prototype.getState = function () {
        return this.state.getValue();
    };
    ReduxTestingStore.prototype.subscribe = function (listener) {
        return this.state.subscribe.call(this.state, listener);
    };
    ReduxTestingStore.prototype.replaceReducer = function () {
    };
    ReduxTestingStore.prototype.dispatch = function (action) {
        return action;
    };
    return ReduxTestingStore;
}());
ReduxTestingStore.decorators = [
    { type: Injectable },
];
ReduxTestingStore.ctorParameters = function () { return []; };
var TestingStateProvider = TestingStateProvider_1 = /** @class */ (function (_super) {
    __extends(TestingStateProvider, _super);
    function TestingStateProvider() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    TestingStateProvider.prototype.getInitialState = function () {
        return TestingStateProvider_1.INITIAL_STATE;
    };
    return TestingStateProvider;
}(ReduxStateProvider));
TestingStateProvider.NAME = 'testing-7c66b613-20bd-4d35-8611-5181ca4a0b72';
TestingStateProvider.INITIAL_STATE = {
    todo: {
        isFetching: false,
        items: ['Item 1', 'Item 2'],
    },
};
TestingStateProvider = TestingStateProvider_1 = __decorate([
    ReduxStateDecoratorForClass({ name: TestingStateProvider_1.NAME })
], TestingStateProvider);
var TestingStateProvider_1;
var ReduxModule = /** @class */ (function () {
    function ReduxModule(injector, reducerProvider, stateDefs) {
        if (stateDefs === void 0) { stateDefs = []; }
        injector.get(ReduxRegistry);
        if (Array.isArray(stateDefs)) {
            stateDefs
                .filter(function (def) { return def && def.provider; })
                .map(function (def) { return injector.get(def.provider); })
                .forEach(function (provider) { return reducerProvider.addStateProvider(provider); });
        }
    }
    ReduxModule.forChild = function (config) {
        if (config === void 0) { config = {}; }
        return {
            ngModule: ReduxModule,
            providers: [
                { provide: ReduxStateDefinitionToken, useValue: config.state || null, multi: true },
            ],
        };
    };
    ReduxModule.forRoot = function (config) {
        if (config === void 0) { config = {}; }
        return {
            ngModule: ReduxModule,
            providers: [
                ReduxReducerProvider,
                ReduxRegistry,
                {
                    provide: ReduxStore,
                    useFactory: config.storeFactory || ReduxModule.defaultStoreFactory,
                    deps: [ReduxReducerProvider]
                },
                { provide: ReduxStateDefinitionToken, useValue: config.state || null, multi: true },
            ],
        };
    };
    ReduxModule.defaultStoreFactory = function (reduxReducerProvider, devMode) {
        if (devMode === void 0) { devMode = isDevMode(); }
        return createStore(reduxReducerProvider.rootReducer, {}, ReduxModule.defaultEnhancerFactory(devMode));
    };
    ReduxModule.defaultEnhancerFactory = function (devMode) {
        if (window['__REDUX_DEVTOOLS_EXTENSION__'] && devMode) {
            return window['__REDUX_DEVTOOLS_EXTENSION__']();
        }
        return ReduxModule.noopEnhancer;
    };
    ReduxModule.noopEnhancer = function (next) {
        return next;
    };
    return ReduxModule;
}());
ReduxModule.decorators = [
    { type: NgModule, args: [{
                declarations: [
                    ReduxSelectPipe,
                ],
                exports: [
                    ReduxSelectPipe,
                ],
                imports: [
                    CommonModule,
                ],
            },] },
];
ReduxModule.ctorParameters = function () { return [
    { type: Injector, },
    { type: ReduxReducerProvider, },
    { type: Array, decorators: [{ type: Optional }, { type: Inject, args: [ReduxStateDefinitionToken,] },] },
]; };
var getActionType = ReduxActionDispatcher.getType;
var dispatch = ReduxActionDispatcher.dispatch;

export { ReduxModule, getActionType, dispatch, ReduxSelect, ReduxReducerProvider, ReduxSelectPipe, ReduxSelector, ReduxStateDefinitionToken, ReduxStateProvider, ReduxStore, ReduxTestingStore, TestingStateProvider, ReduxRegistry };
//# sourceMappingURL=harmowatch-ngx-redux-core.js.map
